@@ -1,7 +1,7 @@
 /************************************************************************
 **
-**  Copyright (C) 2015 Kevin B. Hendricks Stratford, ON Canada 
-**  Copyright (C) 2009, 2010, 2011  Strahinja Markovic  <strahinja.markovic@gmail.com>
+**  Copyright (C) 2015-2021 Kevin B. Hendricks Stratford, ON Canada 
+**  Copyright (C) 2009-2011  Strahinja Markovic  <strahinja.markovic@gmail.com>
 **
 **  This file is part of Sigil.
 **
@@ -25,15 +25,17 @@
 #define CODEVIEWEDITOR_H
 
 #include <QtCore/QList>
-#include <QtCore/QStack>
 #include <QtWidgets/QPlainTextEdit>
 #include <QtGui/QStandardItem>
 #include <QtCore/QUrl>
 
-#include "Misc/CSSInfo.h"
+#include "Parsers/CSSInfo.h"
+#include "Parsers/HTMLStyleInfo.h"
 #include "Misc/PasteTarget.h"
 #include "Misc/SettingsStore.h"
 #include "Misc/Utility.h"
+#include "Widgets/TextDocument.h"
+#include "Parsers/TagLister.h"
 #include "MiscEditors/ClipEditorModel.h"
 #include "MiscEditors/IndexEditorModel.h"
 #include "ViewEditors/ViewEditor.h"
@@ -77,6 +79,8 @@ public:
     CodeViewEditor(HighlighterType highlighter_type, bool check_spelling = false, QWidget *parent = 0);
     ~CodeViewEditor();
 
+    void SetAppearance();
+
     QSize sizeHint() const;
 
     /**
@@ -85,7 +89,7 @@ public:
      *
      * @param document The new text document.
      */
-    void CustomSetDocument(QTextDocument &document);
+    void CustomSetDocument(TextDocument &document);
 
     void DeleteLine();
 
@@ -93,15 +97,22 @@ public:
 
     bool MoveToMarkedText(Searchable::Direction direction, bool wrap);
 
+    bool MoveToSplitText(Searchable::Direction direction, int start, int end);
+
     /**
      * Routines to handle cutting code tags from selected text
      */
     void CutCodeTags();
     bool IsCutCodeTagsAllowed();
 
+    void CutTagPair();
+    bool IsCutTagPairAllowed();
+
     bool TextIsSelected();
     bool TextIsSelectedAndNotInStartOrEndTag();
-    bool TextIsSelectedAndNotContainingTag();
+
+    // No Longer Exists?
+    // bool TextIsSelectedAndNotContainingTag();
 
     QString StripCodeTags(QString text);
 
@@ -126,7 +137,9 @@ public:
     bool InsertHyperlink(const QString &attribute_value);
     bool IsInsertIdAllowed();
     bool IsInsertHyperlinkAllowed();
-    bool InsertTagAttribute(const QString &element_name, const QString &attribute_name, const QString &attribute_value, const QStringList &tag_list, bool ignore_seletion = false);
+    bool InsertTagAttribute(const QString &element_name, const QString &attribute_name, 
+                            const QString &attribute_value, const QStringList &tag_list, 
+                            bool ignore_seletion = false);
 
     /**
     * Splits the section and returns the "upper" content.
@@ -199,6 +212,23 @@ public:
 
     void ScrollToFragment(const QString &fragment);
 
+    // get length of plain text
+    int textLength() const;
+
+    // override and hide the toPlainText() call to prevent
+    // issues with lost non-breaking spaces (improperly
+    // converted to normal spaces)
+    QString toPlainText() const;
+
+    // override the createMimeDataFromSelection() to 
+    // prevent copy and cut from losing nbsp
+    // ala Kovid's solution in calibre PlainTextEdit
+    virtual QMimeData *createMimeDataFromSelection() const;
+
+    // override the insertFromMimeData() to 
+    // attempt to normalize any text to form NFC before paste
+    virtual void insertFromMimeData(const QMimeData* source);
+
     // inherited
     bool IsLoadingFinished();
 
@@ -225,7 +255,8 @@ public:
                   bool misspelled_words = false,
                   bool ignore_selection_offset = false,
                   bool wrap = true,
-                  bool selected_text = false);
+                  bool selected_text = false,
+                  int split_at = -1);
 
     int Count(const QString &search_regex, Searchable::Direction direction, bool wrap, bool selected_text = false);
 
@@ -250,11 +281,13 @@ public:
      */
     void SetDelayedCursorScreenCenteringRequired();
 
-    // inherited
-    QList<ViewEditor::ElementIndex> GetCaretLocation();
+    QString GetCaretElementName();
 
     // inherited
-    void StoreCaretLocationUpdate(const QList<ViewEditor::ElementIndex> &hierarchy);
+    QList<ElementIndex> GetCaretLocation();
+
+    // inherited
+    void StoreCaretLocationUpdate(const QList<ElementIndex> &hierarchy);
 
     // inherited
     bool ExecuteCaretUpdate(bool default_to_top = false);
@@ -309,6 +342,12 @@ public:
      */
     void FormatCSSStyle(const QString &property_name, const QString &property_value);
 
+    bool IsSelectionValid(const QString &text);
+
+    void WrapSelectionInElement(const QString &element, bool unwrap = false);
+
+    void ApplyListToSelection(const QString &element);
+
     void ApplyCaseChangeToSelection(const Utility::Casing &casing);
 
     QString GetAttributeId();
@@ -318,7 +357,10 @@ public:
      * get appropriate attribute_value for attribute_name.
      *
      */
-    QString GetAttribute(const QString &attribute_name, QStringList tag_list = QStringList(), bool must_be_in_attribute = false, bool skip_paired_tags = false);
+    QString GetAttribute(const QString &attribute_name, QStringList tag_list = QStringList(),
+                         bool must_be_in_attribute = false, bool skip_paired_tags = false,
+                         bool must_be_in_body = true);
+
 
     QString SetAttribute(const QString &attribute_name, QStringList tag_list = QStringList(), const QString &attribute_value = QString(), bool must_be_in_attribute = false, bool skip_paired_tags = false);
 
@@ -329,8 +371,11 @@ public:
      * @param attribute_name The name of the attribute to be inserted/replaced.
      * @param attribute_value The new value to be assigned to this attribute.
      */
-    QString ProcessAttribute(const QString &attribute_name, QStringList tag_list = QStringList(), const QString &attribute_value = QString(), bool set_attribute = false , bool must_be_in_attribute = false, bool skip_paired_tags = false);
-
+    QString ProcessAttribute(const QString &attribute_name, QStringList tag_list = QStringList(),
+                             const QString &attribute_value = QString(), bool set_attribute = false,
+                             bool must_be_in_attribute = false, bool skip_paired_tags = false,
+                             bool must_be_in_body = true);
+    
     /**
      * Control whether the Reformat CSS submenu is available on the context menu.
      */
@@ -404,7 +449,10 @@ public slots:
      *
      * @param printer The printer interface to use for printing.
      */
-    void print(QPagedPaintDevice *printer);
+    void print(QPrinter *printer);
+
+    void RenameClassClicked();
+    void RenameClass();
 
     // Implementations for PasteTarget.h
     void PasteText(const QString &text);
@@ -417,10 +465,17 @@ public slots:
     void GoToLinkOrStyle();
 
     bool MarkSelection();
+
     bool ClearMarkedText();
 
+    void SetAppearanceColors();
+
+    void RehighlightDocument();
+
 protected:
 
+    QString cursor_selected_text(const QTextCursor& c) const;
+    
     /**
      * The global event processing function.
      *
@@ -444,6 +499,8 @@ protected:
 
     void mouseReleaseEvent(QMouseEvent *event);
 
+    void mouseDoubleClickEvent(QMouseEvent *event);
+    
     /**
      * Handles the content menu event for the editor.
      *
@@ -476,8 +533,6 @@ private slots:
      */
     void TextChangedFilter();
 
-    void RehighlightDocument();
-
     void PasteClipEntryFromName(const QString &name);
 
     /**
@@ -505,7 +560,7 @@ private slots:
     /**
      * Highlights the line the user is editing.
      */
-    void HighlightCurrentLine();
+    void HighlightCurrentLine(bool highlight_tags=true);
 
     /**
      * Wrapper slot for the Scroll One Line Up shortcut.
@@ -527,6 +582,8 @@ private slots:
 
     void GoToLinkOrStyleAction();
 
+    void ToggleLineWrapMode();
+
     void ReformatCSSMultiLineAction();
     void ReformatCSSSingleLineAction();
 
@@ -538,15 +595,15 @@ private slots:
 private:
     bool IsMarkedText();
 
+    void MaybeRegenerateTagList();
+
+    QString RemoveFirstTag(const QString &text, const QString &tagname);
+    QString RemoveLastTag(const QString &text, const QString &tagname);
+
     QString GetCurrentWordAtCaret(bool select_word);
 
     bool PasteClipEntry(ClipEditorModel::clipEntry *clip);
 
-    /**
-     * Returns the text inside < > if cursor is in < >
-     */
-    QString GetTagText();
-
     /**
      * Resets the currently used font.
      */
@@ -560,8 +617,6 @@ private:
      */
     void UpdateLineNumberAreaFont(const QFont &font);
 
-    void SetAppearanceColors();
-
     /**
      * Executes a centerCursor() call if requested
      * with m_DelayedCursorScreenCenteringRequired.
@@ -594,10 +649,14 @@ private:
 
     void AddReformatCSSContextMenu(QMenu *menu);
 
+    void AddCSSClassContextMenu(QMenu *menu);
+
     void AddReformatHTMLContextMenu(QMenu *menu);
 
     void AddGoToLinkOrStyleContextMenu(QMenu *menu);
 
+    void AddToggleLineWrapModeContextMenu(QMenu *menu);
+
     void AddClipContextMenu(QMenu *menu);
 
     void AddMarkSelectionMenu(QMenu *menu);
@@ -610,46 +669,10 @@ private:
 
     bool InViewableImage();
 
-    /**
-     * An element on the stack when searching for
-     * the current caret location.
-     */
-    struct StackElement {
-        /**
-         * The tag name.
-         */
-        QString name;
+    // Used to convert between Hierarchy and QWedPath used by Preview, Gumbo, and TagLister
+    QList<ElementIndex> ConvertQWebPathToHierarchy(const QString& qwebpath) const;
 
-        /**
-         * The number of child elements
-         * detected for the element, so far.
-         */
-        int num_children;
-    };
-
-    /**
-     * Returns a stack of elements representing the
-     * current location of the caret in the document.
-     *
-     * @param offset The number of characters from document start to the end of
-     *               the start tag of the element the caret is residing in.
-     * @return The element location stack.
-     */
-    QStack<StackElement> GetCaretLocationStack(int offset) const;
-
-    /**
-     * Takes the stack provided by GetCaretLocationStack()
-     * and converts it into the element location hierarchy
-     * used by other ViewEditors.
-     *
-     * @param stack The StackElement stack.
-     * @return The converted ElementIndex hierarchy.
-     */
-    QList<ElementIndex> ConvertStackToHierarchy(const QStack<StackElement> stack) const;
-
-
-    // Used to convert Hierarchy to QWedPath used by BV and Gumbo
-    QString ConvertHierarchyToQWebPath(const QList<ViewEditor::ElementIndex>& hierarchy) const;
+    QString ConvertHierarchyToQWebPath(const QList<ElementIndex>& hierarchy) const;
 
     /**
      * Converts a ViewEditor element hierarchy to a tuple describing necessary caret moves.
@@ -658,7 +681,7 @@ private:
      * @param hierarchy The caret location as ElementIndex hierarchy.
      * @return The info needed to move the caret to the new location.
      */
-    std::tuple<int, int> ConvertHierarchyToCaretMove(const QList<ViewEditor::ElementIndex> &hierarchy) const;
+    std::tuple<int, int> ConvertHierarchyToCaretMove(const QList<ElementIndex> &hierarchy) const;
 
     /**
      * Insert HTML tags around the current selection.
@@ -670,14 +693,14 @@ private:
     /**
      * Is this position within the <body> tag of this text.
      */
-    bool IsPositionInBody(const int &pos = -1, const QString &text = QString());
-    bool IsPositionInTag(const int &pos = -1, const QString &text = QString());
-    bool IsPositionInOpeningTag(const int &pos = -1, const QString &text = QString());
-    bool IsPositionInClosingTag(const int &pos = -1, const QString &text = QString());
-    QString GetOpeningTagName(const int &pos, const QString &text);
-    QString GetClosingTagName(const int &pos, const QString &text);
+    bool IsPositionInBody(int pos);
+    bool IsPositionInTag(int pos);
+    bool IsPositionInOpeningTag(int pos);
+    bool IsPositionInClosingTag(int pos);
+    QString GetOpeningTagName(int pos);
+    QString GetClosingTagName(int pos);
 
-    void FormatSelectionWithinElement(const QString &element_name, const int &previous_tag_index, const QString &text);
+    void FormatSelectionWithinElement(const QString &element_name, int tagno, const QString &text);
 
     void ReplaceTags(const int &opening_tag_start, const int &opening_tag_end, const QString &opening_tag_text,
                      const int &closing_tag_start, const int &closing_tag_end, const QString &closing_tag_text);
@@ -709,13 +732,13 @@ private:
      * Given a list of CSS properties perform any pruning/replacing/adding as necessary to
      * ensure that property_name:property_value is added (or removed if it already exists).
      */
-    void ApplyChangeToProperties(QList<CSSInfo::CSSProperty *> &css_properties, const QString &property_name, const QString &property_value);
+    void ApplyChangeToProperties(QList<HTMLStyleInfo::CSSProperty> &css_properties, const QString &property_name, const QString &property_value);
 
     void ReformatCSS(bool multiple_line_format);
 
     void ReformatHTML(bool all, bool to_valid);
 
-    QStringList GetUnmatchedTagsForBlock(const int &pos, const QString &text) const;
+    QStringList GetUnmatchedTagsForBlock(int pos);
 
     void SelectAndScrollIntoView(int start_position, int end_position, Searchable::Direction direction, bool wrapped);
 
@@ -783,7 +806,7 @@ private:
      * Stores the update for the caret location
      * when switching from BookView to CodeView.
      */
-    QList<ViewEditor::ElementIndex> m_CaretUpdate;
+    QList<ElementIndex> m_CaretUpdate;
 
     /**
      * Whether spell checking is enabled on this view.
@@ -834,6 +857,10 @@ private:
      * Whether spelling highlighting should be reapplied when this tab is next given focus.
      */
     bool m_pendingSpellingHighlighting;
+    QString m_element_name;
+
+    TagLister m_TagList;
+    bool m_regen_taglist;
 };
 
 #endif // CODEVIEWEDITOR_H
